@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 #
 # This file is part of urlwatch (https://thp.io/2008/urlwatch/).
-# Copyright (c) 2008-2022 Thomas Perl <m@thp.io>
+# Copyright (c) 2008-2024 Thomas Perl <m@thp.io>
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -49,21 +49,25 @@ class Mailer(object):
     def send(self, msg):
         raise NotImplementedError
 
-    def msg_plain(self, from_email, to_email, subject, body):
+    def msg_plain(self, from_email, to_email, reply_to_email, subject, body):
         msg = email.mime.text.MIMEText(body, 'plain', 'utf-8')
         msg['Subject'] = subject
         msg['From'] = from_email
         msg['To'] = to_email
         msg['Date'] = email.utils.formatdate()
+        if reply_to_email:
+            msg['Reply-To'] = reply_to_email
 
         return msg
 
-    def msg_html(self, from_email, to_email, subject, body_text, body_html):
+    def msg_html(self, from_email, to_email, reply_to_email, subject, body_text, body_html):
         msg = email.mime.multipart.MIMEMultipart('alternative')
         msg['Subject'] = subject
         msg['From'] = from_email
         msg['To'] = to_email
         msg['Date'] = email.utils.formatdate()
+        if reply_to_email:
+            msg['Reply-To'] = reply_to_email
 
         msg.attach(email.mime.text.MIMEText(body_text, 'plain', 'utf-8'))
         msg.attach(email.mime.text.MIMEText(body_html, 'html', 'utf-8'))
@@ -99,7 +103,7 @@ class SMTPMailer(Mailer):
                 raise ValueError('SMTP auth is enabled, but insecure_password is not set and keyring is not available')
             s.login(self.smtp_user, passwd)
 
-        s.sendmail(msg['From'], msg['To'].split(','), msg.as_string())
+        s.sendmail(msg['From'], msg['To'].split(','), msg.as_string(maxheaderlen=78))
         s.quit()
 
 
@@ -108,11 +112,11 @@ class SendmailMailer(Mailer):
         self.sendmail_path = sendmail_path
 
     def send(self, msg):
-        p = subprocess.Popen([self.sendmail_path, '-oi', '-f', msg['From'], msg['To']],
+        p = subprocess.Popen([self.sendmail_path, '-oi', '-f', msg['From']] + msg['To'].split(','),
                              stdin=subprocess.PIPE,
                              stderr=subprocess.PIPE,
                              universal_newlines=True)
-        result = p.communicate(msg.as_string())
+        result = p.communicate(msg.as_string(maxheaderlen=78))
         if p.returncode:
             logger.error('Sendmail failed with {result}'.format(result=result))
 
